/* $Id: imgmem.cpp,v 1.1.1.1 1995/07/31 13:53:14 cvs Exp $
 *
 * Copyright (c) 1996 University of Alberta
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See COPYING file for more information.
 */

//  Object procedures for manipulating digital Images in memory

#include <owl\owlcore.h>
#include <alloc.h>
#pragma hdrstop

#include "image.h"


// Image methods

void Image::SetPalette (const TPalette &APalette)
  {
  if (palette) delete palette;
  palette = new TPalette(APalette);
  }

int Image::CopyFrom(Image *AImage)
  {
  int w, h, y, i;

  const TPalette *pal = AImage->GetPalette();
  if ((!palette) && (pal)) SetPalette(*pal);

  AImage->GetImageSize (w, h);
  if (!SetImageSize (w, h)) return 0;
  unsigned char * line = (unsigned char *) malloc (w*3);
  if (!line) return 0;

  if (AImage->GetOrientation()&1)
	for (y=h-1; y>=0; y--)
	  {
	  if (StatusFunc) StatusFunc ((long)100*(h-y-1)/h);
	  AImage->ReadLineRGB (y, line);
	  WriteLineRGB (y, line);
	  }
  else for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	AImage->ReadLineRGB (y, line);
	WriteLineRGB (y, line);
	}
  free (line);
  return 1;
  }


// MemoryImage methods

void MemoryImage::GetPixelRGB(int x, int y, int& r, int& g, int& b)
  {
  long c;
  r=g=b=0;
  switch (imageType) {
	case (TRUECOLOR) :
	  GetPixel (x, y, c);
	  if (blueBits)
		{
		if (blueBits==8) b = (unsigned char) c;
		  else {
			b = (((unsigned char) c) && ((0xFF<<blueBits) ^ 0xFF)) << (8-blueBits);
			b += b >> blueBits;
			}
		c >>= blueBits;
		}
	  if (greenBits)
		{
		if (greenBits==8) g = (unsigned char) c;
		  else {
			g = (((unsigned char) c) && ((0xFF<<greenBits) ^ 0xFF)) << (8-greenBits);
			g += g >> greenBits;
			}
		c >>= greenBits;
		}
	  if (redBits)
		{
		if (redBits==8) r = (unsigned char) c;
		  else {
			r = (((unsigned char) c) && ((0xFF<<redBits) ^ 0xFF)) << (8-redBits);
			r += r >> redBits;
			}
		}
	  break;
	case (GRAYSCALE) :
	  GetPixel (x, y, c);
	  c <<= (8-bitsPerPixel);
	  r = g = b = c + (c >> bitsPerPixel);
	  break;
	}
  }

void MemoryImage::PutPixelRGB(int x, int y, int r, int g, int b)
  {
  long c;
  switch (imageType) {
	case (TRUECOLOR) :
	  if (redBits != 8)   r >>= (8-redBits);
	  if (greenBits != 8) g >>= (8-greenBits);
	  if (blueBits != 8)  b >>= (8-blueBits);
	  c = (((long)(r << greenBits) + g) << blueBits) + b;
	  PutPixel (x, y, c);
	  break;
	case (GRAYSCALE) :
	  c = ((r+g+b)/3) >> (8-bitsPerPixel);
	  PutPixel (x, y, c);
	  break;
	}
  }

void MemoryImage::GetPixelGray (int x, int y, int& gray)
  {
  int r, g, b;
  long c;
  switch (imageType) {
	case (TRUECOLOR) :
	  GetPixelTC (x, y, r, g, b);
	  gray = (r+g+b) / 3;
	  break;
	case (GRAYSCALE) :
	  GetPixel (x, y, c);
	  gray = c << (8-bitsPerPixel);
	  gray += gray >> bitsPerPixel;
	  break;
	}
  }

void MemoryImage::PutPixelGray (int x, int y, int gray)
  {
  switch (imageType) {
	case (TRUECOLOR) :
	  PutPixelTC (x, y, gray, gray, gray);
	  break;
	case (GRAYSCALE) :
	  gray >>= (8-bitsPerPixel);
	  PutPixel (x, y, gray);
	  break;
	}
  }

void MemoryImage::GetPixelCLUT (int x, int y, int& i)
  {
  long c;
  GetPixel (x, y, c);
  i = LOWORD(c);
  }

void MemoryImage::PutPixelCLUT (int x, int y, int i)
  {
  long c = i;
  PutPixel (x, y, c);
  }

void MemoryImage::GetPixel (int x, int y, long& c)
  {
  long a = h-1-y;  a = a*w+x;
  switch (bitsPerPixel) {
	case 1 :
	case 2 :
	case 4 :
	  c = image[(a*bitsPerPixel)>>3];
	  c >>= (x*bitsPerPixel) & 0x07;
	  c &= (0xFFFFFFFF << bitsPerPixel) ^ 0xFFFFFFFF;
	  break;
	case 8 :
	  c = image[a];
	  break;
	case 16 :
	  c = *((unsigned int *)&image[a<<1]);
	  break;
	case 24 :
	  c = *((long *)&image[a*3]) & 0xFFFFFF;
	  break;
	}
  }

void MemoryImage::PutPixel (int x, int y, long c)
  {
  long a=h-1-y;  a = a*w+x;
  switch (bitsPerPixel) {
	 case 1 :
	 case 2 :
	 case 4 :
		{
		long v = image[(a*bitsPerPixel)>>3];
		v &= (((0xFFFFFFFF << bitsPerPixel) ^ 0xFFFFFFFF) << ((a*bitsPerPixel) & 0x07))^0xFFFFFFFF;
		image[(a*bitsPerPixel)>>3] = v | (c << ((a*bitsPerPixel) & 0x07));
		}
		break;
	case 8 :
	  image[a] = c;
	  break;
	case 16 :
	  image[a<<1] = c & 0xFF;
	  image[(a<<1)+1] = c >> 8;
	  break;
	case 24 :
	  image[a*3] = c & 0xFF;
	  image[(a*3)+1] = c >> 8;
	  image[(a*3)+2] = c >> 16;
	  break;
	}
  }

void MemoryImage::PutPixel (tColorPlane cp, int x, int y, long c)
  {
  int r, g, b;
  long v, mask = 0xFFFFFFFF;
  switch (cp) {
	case (blue) :
	  GetPixel (x, y, v);
	  if (blueBits == 8) v = (v&0xFFFFFF00) + (c&0x000000FF);
		else {
		  mask <<= blueBits;
		  c >>= (8-blueBits);
		  v = (v & mask) + c;
		  }
	  break;
	case (green) :
	  GetPixel (x, y, v);
	  mask <<= greenBits;
	  mask = (mask^0xFFFFFFFF) << blueBits;
	  mask ^= 0xFFFFFFFF;
	  c >>= (8-greenBits);
	  c <<= blueBits;
	  v = (v & mask) + c;
	  break;
	case (red) :
	  GetPixel (x, y, v);
	  mask <<= redBits;
	  mask = (mask^0xFFFFFFFF) << (blueBits + greenBits);
	  mask ^= 0xFFFFFFFF;
	  c >>= (8-redBits);
	  c <<= blueBits+greenBits;
	  v = (v & mask) + c;
	  break;
	case (gray) :
	  v = c >> (8-bitsPerPixel);
	  break;
	}
  PutPixel (x, y, v);
  }

void MemoryImage::ReadLineRGB (int y, unsigned char *buf)
  {
  int x, r, g, b, i=0;

  switch (imageType) {
	case (TRUECOLOR) :
	case (GRAYSCALE) :
	  for (x=0; x<w; x++)
		GetPixelTC (x, y, r, g, b),
		buf[i++]=r, buf[i++]=g, buf[i++]=b;
	  break;
	}
  }

void MemoryImage::WriteLineRGB (int y, unsigned char *buf)
  {
  int x=0, i;

  switch (imageType) {
	case (TRUECOLOR) :
	case (GRAYSCALE) :
	  for (i=0; i<w*3; i+=3)
		PutPixelTC (x++, y, buf[i], buf[i+1], buf[i+2]);
	  break;
	}
  }

void MemoryImage::ClearImage (tColorPlane tc, int color)
  {
  int x, y;
  for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  PutPixel (tc, x, y, color);
	}
  }

void MemoryImage::AcquireImage () {}

void MemoryImage::SetImageType (int type)
  {
  if (((imageType==GRAYSCALE) || (imageType==CLUT)) &&
	  ((type==GRAYSCALE) || (type==CLUT))) imageType = type;
  }

void MemoryImage::Smooth (tColorPlane cp, int matrixSize, int x1, int y1, int x2, int y2)
  {
  }

void MemoryImage::Smooth (tColorPlane cp, int matrixSize)
  {
  int x, y, i, j, r, g, b, lastline, matrixArm, n, endx, endy;
  unsigned long v, add;
  unsigned char* lines = (unsigned char*) malloc (w*matrixSize);
  if (!lines) return;
  lastline=matrixSize-1;  matrixArm = (matrixSize-1)/2;

  switch (cp) {
	case (red) :
	  for (y=0; y<matrixSize; y++)
		for (x=0; x<w; x++)
		  {
		  GetPixelTC (x, y, r, g, b);
		  lines[y*w+x] = r;
		  }
	  break;
	case (green) :
	  for (y=0; y<matrixSize; y++)
		for (x=0; x<w; x++)
		  {
		  GetPixelTC (x, y, r, g, b);
		  lines[y*w+x] = g;
		  }
	  break;
	case (blue) :
	  for (y=0; y<matrixSize; y++)
		for (x=0; x<w; x++)
		  {
		  GetPixelTC (x, y, r, g, b);
		  lines[y*w+x] = b;
		  }
	  break;
	case (gray) :
	  for (y=0; y<matrixSize; y++)
		for (x=0; x<w; x++)
		  {
		  GetPixelGray (x, y, g);
		  lines[y*w+x] = g;
		  }
	}
  for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	if (((y+matrixArm) > lastline) && (lastline < h-1))
	  {
	  lastline++;
	  switch (cp) {
		case (red) :
		  for (x=0; x<w; x++)
			{
			GetPixelTC (x, lastline, r, g, b);
			lines[(lastline%matrixSize)*w+x] = r;
			}
		  break;
		case (green) :
		  for (x=0; x<w; x++)
			{
			GetPixelTC (x, lastline, r, g, b);
			lines[(lastline%matrixSize)*w+x] = g;
			}
		  break;
		case (blue) :
		  for (x=0; x<w; x++)
			{
			GetPixelTC (x, lastline, r, g, b);
			lines[(lastline%matrixSize)*w+x] = b;
			}
		  break;
		case (gray) :
		  for (x=0; x<w; x++)
			{
			GetPixelGray (x, lastline, g);
			lines[(lastline%matrixSize)*w+x] = g;
			}
		  break;
		}
	  }
	if (matrixSize != 3)
	  for (x=0; x<w; x++)
		{
		v = n = 0;
		for (j=0, (y<matrixArm ? endy=matrixSize+y-matrixArm : endy=matrixSize); j<endy; j++)
		  for ((x<matrixArm ? i=0 : i=x-matrixArm), (x+matrixArm>=w ? endx=w : endx=x+matrixArm+1); i<endx; i++)
			{
			add = lines[j*w+i];
			if (i==x) add << 1, n++;
			if (j==y) add << 1, n++;
			if ((i==x) && (j==y)) n++;
			v += add;  n++;
			}
		PutPixel (cp, x, y, v/n);
		}
	  else {
		if (y != 0)
		  {
		  v = lines[0]   + lines[1]   + lines[w] +
		  lines[w+1] + lines[w+w] + lines [w+w+1];
		  PutPixel (cp, 0, y, v/6);
		  for (x=1; x<w-1; x++)
			{
			v = lines[x-1]     + lines[x]     + lines[x+1] +
				lines[w+x-1]   + lines[w+x]   + lines[w+x+1] +
				lines[w+w+x-1] + lines[w+w+x] + lines [w+w+x+1];
			PutPixel (cp, x, y, (v>>3) - (v>>6) + (v>>9));
			}
		  v = lines[w-2]   + lines[w-1]     + lines[w+w-2] +
			  lines[w+w-1] + lines[w+w+w-2] + lines[w+w+w-1];
		  PutPixel (cp, w-1, y, v/6);
		  }
		else {
		  v = lines[0]   + lines[1]   + lines[w] + lines[w+1];
		  PutPixel (cp, 0, y, v/4);
		  for (x=1; x<w-1; x++)
			{
			v = lines[x-1]     + lines[x]     + lines[x+1] +
				lines[w+x-1]   + lines[w+x]   + lines[w+x+1];
			PutPixel (cp, x, y, v/6);
			}
		  v = lines[w-2] + lines[w-1] + lines[w+w-2] + lines[w+w-1];
		  PutPixel (cp, w-1, y, v/4);
		  }
		}
	}
  }

void MemoryImage::MatrixFilter (tColorPlane cp, int matrix[], int matrixSize, int x1, int y1, int x2, int y2)
  {
  }

// maximum matrixSize = 21
void MemoryImage::MatrixFilter (tColorPlane cp, int matrix[], int matrixSize)
  {
  int x, y, i, j, r, g, b, lastline, matrixArm, n, endx, endy;
  long v;
  unsigned char* lines = (unsigned char*) malloc (w*matrixSize);
  int linenum[21];
  if (!lines) return;
  lastline=matrixSize-1;  matrixArm = (matrixSize-1)/2;

  for (y=0; y<matrixSize; y++) linenum[y] = y;
  switch (cp) {
	case (red) :
	  for (y=0; y<matrixSize; y++)
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  lines[y*w+x] = r;
	  };
	  break;
	case (green) :
	  for (y=0; y<matrixSize; y++)
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  lines[y*w+x] = g;
	  };
	  break;
	case (blue) :
	  for (y=0; y<matrixSize; y++)
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  lines[y*w+x] = b;
	  };
      break;
    case (gray) :
      for (y=0; y<matrixSize; y++)
	for (x=0; x<w; x++)
	  {
	  GetPixelGray (x, y, g);
	  lines[y*w+x] = g;
	  };
    };
  for (y=1; y<h-1; y++)
    {
    if (StatusFunc) StatusFunc ((long)100*y/h);
    if (((y+matrixArm) > lastline) && (lastline < h-1))
      {
      for (i=0; i<matrixSize; i++) linenum[i] = (linenum[i]+1)%matrixSize;
      lastline++;
      switch (cp) {
	case (red) :
	  for (x=0; x<w; x++)
	    {
	    GetPixelTC (x, lastline, r, g, b);
	    lines[(lastline%matrixSize)*w+x] = r;
	    };
	  break;
	case (green) :
	  for (x=0; x<w; x++)
	    {
	    GetPixelTC (x, lastline, r, g, b);
	    lines[(lastline%matrixSize)*w+x] = g;
	    };
	  break;
	case (blue) :
	  for (x=0; x<w; x++)
		{
	    GetPixelTC (x, lastline, r, g, b);
	    lines[(lastline%matrixSize)*w+x] = b;
	    };
	  break;
	case (gray) :
	  for (x=0; x<w; x++)
	    {
	    GetPixelGray (x, lastline, g);
	    lines[(lastline%matrixSize)*w+x] = g;
	    };
	  break;
	};
      };
    for (x=1; x<w-1; x++)
      {
      v = n = 0;
      for (j=0; j<2; j++)
	for (i=0; i<2; i++)
	  v += lines[linenum[j]*w+i+x-1] * matrix[j*matrixSize+i],
	  n += matrix[j*matrixSize+i];
      if (n>0)  v/=n;
      if (n==0) v+=128;
      if (n<0)  v/=-n;
      if (v<0)   v=0;
      if (v>255) v=255;
      PutPixel (cp, x, y, v);
	  }
    };
  };

void MemoryImage::AddImage (class MemoryImage *Aimage)
  {
  int x, y, p1, p2;
  switch (imageType) {
	case (GRAYSCALE) :
	  for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  Aimage-> GetPixelGray (x, y, p1);
	  GetPixelGray (x, y, p2);  p1 += p2;
	  if (p1 > 255) p1 = 255;
	  PutPixelGray (x, y, p1);
	  };
	};
	  break;
	};
  };

void MemoryImage::AverageImage (class MemoryImage *Aimage)
  {
  int x, y, p1, p2;
  switch (imageType) {
	case (GRAYSCALE) :
	  for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  Aimage-> GetPixelGray (x, y, p1);
	  GetPixelGray (x, y, p2);
	  PutPixelGray (x, y, (p1+p2)/2);
	  };
	};
	  break;
	};
  };

void MemoryImage::SubtractImage (class MemoryImage *Aimage)
  {
  int x, y, p1, p2;
  switch (imageType) {
	case (GRAYSCALE) :
	  for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  Aimage-> GetPixelGray (x, y, p1);
	  GetPixelGray (x, y, p2);
	  PutPixelGray (x, y, ((p2-p1)/2)+128);
	  };
	};
	  break;
	};
  };

void MemoryImage::SubtractImageReverse (class MemoryImage *Aimage)
  {
  int x, y, p1, p2;
  switch (imageType) {
	case (GRAYSCALE) :
	  for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  Aimage-> GetPixelGray (x, y, p1);
	  GetPixelGray (x, y, p2);
	  PutPixelGray (x, y, ((p1-p2)/2)+128);
	  };
	};
      break;
    };
  };

void MemoryImage::DivideContrast (tColorPlane cp, int factor)
  {
  int x, y, c, r, g, b;

  switch (cp) {
    case (gray) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelGray (x, y, c);
	  PutPixelGray (x, y, c/factor);
	  };
	};
      break;
    case (red) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  PutPixelTC (x, y, r/factor, g, b);
	  };
	};
      break;
	case (green) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  PutPixelTC (x, y, r, g/factor, b);
	  };
	};
      break;
    case (blue) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  PutPixelTC (x, y, r, g, b/factor);
	  };
	};
      break;
    };
  };

void MemoryImage::MultiplyContrast (tColorPlane cp, int factor)
  {
  int x, y, c, r, g, b;

  switch (cp) {
    case (gray) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelGray (x, y, c);  c *= factor;
	  if (c>255) c=255;
	  PutPixelGray (x, y, c);
	  };
	};
      break;
    case (red) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);  r *= factor;
	  if (r>255) r=255;
	  PutPixelTC (x, y, r, g, b);
	  };
	};
      break;
	case (green) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);  g *= factor;
	  if (g>255) g=255;
	  PutPixelTC (x, y, r, g, b);
	  };
	};
      break;
    case (blue) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);  b *= factor;
	  if (b>255) b=255;
	  PutPixelTC (x, y, r, g, b);
	  };
	};
      break;
    };
  };

void MemoryImage::ShiftIntensity (tColorPlane cp, int cnst)
  {
  int x, y, c, r, g, b;

  switch (cp) {
    case (gray) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelGray (x, y, c);  c += cnst;
	  if (c<0) c=0;
	  if (c>255) c=255;
	  PutPixelGray (x, y, c);
	  };
	};
      break;
    case (red) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);  r += cnst;
	  if (r<0) r=0;
	  if (r>255) r=255;
	  PutPixelTC (x, y, r, g, b);
	  };
	};
      break;
    case (green) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);  g += cnst;
	  if (g<0) g=0;
	  if (g>255) g=255;
	  PutPixelTC (x, y, r, g, b);
	  };
	};
      break;
    case (blue) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);  b += cnst;
	  if (b<0) b=0;
	  if (b>255) b=255;
	  PutPixelTC (x, y, r, g, b);
	  };
	};
      break;
    };
  };

void MemoryImage::EnhanceContrast (tColorPlane cp, int low, int high)
  {
  int x, y, r, g, b, c;

  switch (cp) {
    case (gray) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelGray (x, y, c);
	  if (c<low) c=low; if (c>high) c = high;
	  PutPixelGray (x, y, (c-low)* (long)255/(high-low));
	  };
	};
      break;
    case (red) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  if (r<low) r=low; if (r>high) r = high;
	  PutPixelTC (x, y, (r-low)* (long)255 / (high-low), g, b);
	  };
	};
      break;
    case (green) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  if (g<low) g=low; if (g>high) g = high;
	  PutPixelTC (x, y, r, (g-low)* (long)255/(high-low), b);
	  };
	};
      break;
    case (blue) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  GetPixelTC (x, y, r, g, b);
	  if (b<low) b=low; if (b>high) b = high;
	  PutPixelTC (x, y, r, g, (b-low)* (long)255/(high-low));
	  };
	};
      break;
    };
  };

void MemoryImage::ThresholdImage (tColorPlane cp, int thres, int low, int high)
  {
  int x, y, r, g, b, c;

  switch (cp) {
    case (gray)  :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  GetPixelGray (x, y, c),
	  PutPixelGray (x, y, ( (c<thres) ? (low!=-1 ? low : c) : (high!=-1 ? high : c)) );
	};
      break;
    case (red)   :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  GetPixelTC (x, y, r, g, b),
	  PutPixelTC (x, y, ((r<thres) ? (low!=-1 ? low : r) : (high!=-1 ? high : r)), g, b);
	};
      break;
    case (green) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  GetPixelTC (x, y, r, g, b),
	  PutPixelTC (x, y, r, ((g<thres) ? (low!=-1 ? low : g) : (high!=-1 ? high : g)), b);
	};
      break;
    case (blue)  :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  GetPixelTC (x, y, r, g, b),
	  PutPixelTC (x, y, r, g, ((b<thres) ? (low!=-1 ? low : b) : (high!=-1 ? high : b)));
	};
      break;
    };
  };

void MemoryImage::ReverseImage (tColorPlane cp)
  {
  int x, y, r, g, b, c;

  switch (cp) {
    case (gray) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  GetPixelGray (x, y, c),
	  PutPixelGray (x, y, 255-c);
	};
      break;
    case (red) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  GetPixelTC (x, y, r, g, b),
	  PutPixelTC (x, y, 255-r, g, b);
	};
      break;
    case (green) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  GetPixelTC (x, y, r, g, b),
	  PutPixelTC (x, y, r, 255-g, b);
	};
      break;
    case (blue) :
      for (y=0; y<h; y++)
	{
	if (StatusFunc) StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  GetPixelTC (x, y, r, g, b),
	  PutPixelTC (x, y, r, g, 255-b);
	};
      break;
    };
  };

void MemoryImage::DrawContours (tColorPlane cp, int num, int color)
  {
  int x, y, r, g, b, lev, lev2, div=256/num;
  for (y=0; y<h-1; y++)
    {
    if (StatusFunc) StatusFunc ((long)100*y/h);
    for (x=1; x<w-1; x++)
      {
      GetPixelTC (x, y, r, g, b);  lev = g / div;
      GetPixelTC (x+1, y+1, r, g, b);  lev2 = g/div;
//      if (lev2!=lev) PutPixelTC (x, y, color, color, color);
	  GetPixelTC (x-1, y+1, r, g, b);  lev2 = g/div;
//      if (lev2!=lev) PutPixelTC (x, y, color, color, color);
      GetPixelTC (x, y+1, r, g, b);  lev2 = g/div;
      if (lev2!=lev) PutPixelTC (x, y, color, color, color);
      GetPixelTC (x+1, y, r, g, b);  lev2 = g/div;
      if (lev2!=lev) PutPixelTC (x, y, color, color, color);
      };
    };
  }

void MemoryImage::DrawContours (tColorPlane cp, int num, int color, class MemoryImage *AImage)
  {
  int x, y, r, g, b, lev, lev2, div=256/num;
  for (y=0; y<h-1; y++)
    {
    if (StatusFunc) StatusFunc ((long)100*y/h);
    for (x=1; x<w-1; x++)
      {
      GetPixelTC (x, y, r, g, b);  lev = g / div;
      GetPixelTC (x+1, y+1, r, g, b);  lev2 = g/div;
      if (lev2!=lev) AImage->PutPixelTC (x, y, color, color, color);
      GetPixelTC (x-1, y+1, r, g, b);  lev2 = g/div;
      if (lev2!=lev) AImage->PutPixelTC (x, y, color, color, color);
      GetPixelTC (x, y+1, r, g, b);  lev2 = g/div;
      if (lev2!=lev) AImage->PutPixelTC (x, y, color, color, color);
      GetPixelTC (x+1, y, r, g, b);  lev2 = g/div;
      if (lev2!=lev) AImage->PutPixelTC (x, y, color, color, color);
      };
	};
  };

void MemoryImage::CalculateHistogram ()
  {
  if (!histogram) histogram = new RGBHistogram;
  setmem (histogram, sizeof(*histogram), 0);

  int x, y, r, g, b;
  int c;

  switch (imageType) {
	case (TRUECOLOR) :
	  for (y=0; y<h; y++)
		{
		if (StatusFunc) StatusFunc ((long)100*y/h);
		for (x=0; x<w; x++)
		  {
		  GetPixelTC (x, y, r, g, b);
		  histogram->red[r]++;
		  histogram->green[g]++;
		  histogram->blue[b]++;
		  }
		}
	  break;
	case (GRAYSCALE) :
	  for (y=0; y<h; y++)
		{
		if (StatusFunc) StatusFunc ((long)100*y/h);
		for (x=0; x<w; x++)
		  {
		  GetPixelGray (x, y, c);
		  histogram->red[c]++;
		  }
		}
	  break;
	}
  }

void MemoryImage::SmoothHistogram ()
  {
  if (histogram)
	{
	int i;
	for (i=1; i<255; i++)
	  {
	  histogram->red[i] = (histogram->red[i-1] +
		histogram->red[i] + histogram->red[i+1]) / 3;
	  histogram->green[i] = (histogram->green[i-1] +
		histogram->green[i] + histogram->green[i+1]) / 3;
	  histogram->blue[i] = (histogram->blue[i-1] +
		histogram->blue[i] + histogram->blue[i+1]) / 3;
	  }
	histogram->red[0] = histogram->red[1];
	histogram->red[255] = histogram->red[254];
	histogram->green[0] = histogram->green[1];
	histogram->green[255] = histogram->green[254];
	histogram->blue[0] = histogram->blue[1];
	histogram->blue[255] = histogram->blue[254];
	}
  }

long MemoryImage::GetHistogramValue (tColorPlane cp, int intensity)
  {
  switch (cp) {
	case (gray)  : return (histogram->red[intensity]);
	case (red)   : return (histogram->red[intensity]);
	case (green) : return (histogram->green[intensity]);
	case (blue)  : return (histogram->blue[intensity]);
	}
  return (-1);
  }

int MemoryImage::GetMinimumIntensity (tColorPlane cp)
  {
  int i = 0;
  switch (cp) {
	case (gray)  :
	  while ((histogram->red[i] == 0) && (i<256)) i++;
	  break;
	case (red)   :
	  while ((histogram->red[i] == 0) && (i<256)) i++;
	  break;
	case (green) :
	  while ((histogram->green[i] == 0) && (i<256)) i++;
	  break;
	case (blue)  :
	  while ((histogram->green[i] == 0) && (i<256)) i++;
	  break;
	}
  if (i <= 255) return (i);
  return (-1);
  }

int MemoryImage::GetMaximumIntensity (tColorPlane cp)
  {
  int i=255;
  switch (cp) {
	case (gray)  :
	  while ((histogram->red[i] == 0) && (i>=0)) i--;
	  break;
	case (red)   :
	  while ((histogram->red[i] == 0) && (i>=0)) i--;
	  break;
	case (green) :
	  while ((histogram->green[i] == 0) && (i>=0)) i--;
	  break;
	case (blue)  :
	  while ((histogram->green[i] == 0) && (i>=0)) i--;
	  break;
	}
  if (i >= 0) return (i);
  return (-1);
  }

int MemoryImage::GetBackgroundIntensity (tColorPlane cp)
  {
  long max=0;
  int mi=-1, i;
  switch (cp) {
	case (gray)  :
	  for (i=0; i<256; i++)
		if (histogram->red[i] > max)
		  max=histogram->red[i], mi = i;
	  break;
	case (red)   :
	  for (i=0; i<256; i++)
		if (histogram->red[i] > max)
		  max=histogram->red[i], mi = i;
	  break;
	case (green) :
	  for (i=0; i<256; i++)
		if (histogram->green[i] > max)
		  max=histogram->green[i], mi = i;
	  break;
	case (blue)  :
	  for (i=0; i<256; i++)
		if (histogram->blue[i] > max)
		  max=histogram->blue[i], mi = i;
	  break;
	}
  if (mi == -1) return (-1);
  return (mi);
  }

