/* $Id: palette.cpp,v 1.1.1.1 1994/12/25 21:45:58 cvs Exp $
 *
 * Copyright (c) 1996 University of Alberta
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See COPYING file for more information.
 */

//  Object procedures for manipulating palettes

#include <owl\owlcore.h>
#pragma hdrstop

#include "image.h"

// RGBPalette methods

RGBPalette::RGBPalette ()
  {
  number = 0;
  int i;
  for (i=0; i<256; i++)
    palette[i].red = palette[i].green = palette[i].blue = 0;
  }

RGBPalette::~RGBPalette ()
  {
  }

void RGBPalette::SetEntry (int index, int r, int g, int b)
  {
  if (index > 256) return;
  if (index >= number) number = index+1;
  palette[index].red = r;
  palette[index].green = g;
  palette[index].blue = b;
  }

void RGBPalette::SetEntry(int index, RGBQUAD color)
  {
  if (index > 256) return;
  if (index >= number) number = index+1;
  palette[index].red = color.rgbRed;
  palette[index].green = color.rgbGreen;
  palette[index].blue = color.rgbBlue;
  }


void RGBPalette::GetEntry (int index, int &r, int &g, int &b)
  {
  if (index >= number)
    {
    r=g=b=0;
    return;
    }
  r = palette[index].red;
  g = palette[index].green;
  b = palette[index].blue;
  }

RGBQUAD RGBPalette::GetEntry(int index)
  {
  RGBQUAD result;
  result.rgbReserved=result.rgbRed=result.rgbGreen=result.rgbBlue=0;
  if (index < number)
    {
    result.rgbRed = palette[index].red;
    result.rgbGreen = palette[index].green;
    result.rgbBlue = palette[index].blue;
    }
  return result;
  }

int RGBPalette::GetIndex (int r, int g, int b)
  {
  int bestindex=0;
  long mindist = 1000000, d;
  int i, j;
  for (i=0; i<number; i++)
    {
    j = palette[i].red-r;    d = j*j;
    j = palette[i].green-g;  d += j*j;
    j = palette[i].blue-b;   d += j*j;
    if (d < mindist) mindist=d, bestindex = i;
    };
  return bestindex;
  }

