/* $Id: oldcart.cpp,v 1.1.1.1 1994/02/12 17:39:12 cvs Exp $
 *
 * Copyright (c) 1994 University of Alberta
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See COPYING file for more information.
 */

#define WIN31
#include <image.h>
#include <owl.h>

#define id_MAINMENU "MAINMENU"
#define id_SELECTLINE "SELECTLINE"
#define CM_ENHANCE 101
#define CM_ANALIZE 102
#define CM_OPEN    103
#define CM_BOUNDRY 104
#define CM_BATCH   105

#define variation 	8
#define blocksize 	50
#define nextBlockSize	40
#define edgesize 	10
#define thres 		96
#define countRange	64

#define imageXofs 0
#define imageYofs 0

#define maxpoints 150

struct point {
	 int x, y;
         };

struct tline {
	 int numpoints;
	 point points[maxpoints];
         };

struct selectlinemsg {
	 int top, s_i, i_d, deep, left, right;
         };

// classes

class TApp : public TApplication
  {
  public:
	 TApp (LPSTR AName, HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
		:TApplication (AName, hInstance, hPrevInstance, lpCmdLine, nCmdShow) {};
    virtual void InitMainWindow ();
  };

class TDisplayWindow : public TWindow
  {
  protected:
    HDC memDC;
    HPALETTE pal;
    HBITMAP oldBM;
	 int displaying;
    int w, h;
    HDC dragDC;
    tline topline, rightline, leftline, boneline, s_iline, i_dline;
    tline curline;
  public:
	 TDisplayWindow (PTWindowsObject AParent);
    virtual ~TDisplayWindow ();
    virtual int CanClose ();
    virtual void Circle (int x, int y);
	 virtual void Display ();
	 virtual void Paint (HDC PaintDC, PAINTSTRUCT _FAR & PaintInfo);
	 virtual void Paint (void *, void *)  { };
	 virtual void SaveLines ();
    virtual void SetPalette (HPALETTE p);

	 virtual void WMLButtonDown (RTMessage Msg) = [WM_FIRST + WM_LBUTTONDOWN];
	 virtual void WMMouseMove (RTMessage Msg)   = [WM_FIRST + WM_MOUSEMOVE];
	 virtual void WMLButtonUp (RTMessage Msg)   = [WM_FIRST + WM_LBUTTONUP];
  };

class TMainWindow : public TWindow
  {
  protected:
    HPALETTE pal;
	 PTStatic FuncText;
    TDisplayWindow *DisplayWindow;
    int busy, numcells, wantToQuit;
  public:
	 TMainWindow (PTWindowsObject AParent);
    virtual int CanClose ();
    virtual void ClearDisplay ();

    virtual void CMAnalize (RTMessage Msg) = [CM_FIRST + CM_ANALIZE];
    virtual void CMBatch (RTMessage Msg) = [CM_FIRST + CM_BATCH];
    virtual void CMBoundry (RTMessage Msg) = [CM_FIRST + CM_BOUNDRY];
    virtual void CMEnhance (RTMessage Msg) = [CM_FIRST + CM_ENHANCE];
	 virtual void CMOpen (RTMessage Msg) = [CM_FIRST + CM_OPEN];

    virtual void Display ();
    virtual void FindMax (int x1, int y1, int x2, int y2);
    virtual void FindObject (int x, int y, int c);
    virtual void MakePalette();
	 virtual void Paint (HDC PaintDC, PAINTSTRUCT _FAR &PaintInfo);
	 virtual void Paint (void *, void *)  { };
  };

// Global Variables

char ORGname[64] = "";
char ENHname[64];
char DATname[64];
char LINname[64];
char LISTnames[64][40];
int nNames;
MemoryImage *image;
MemoryImage *mask;
ofstream *f;

PTStatic StatText;
int LastPercent;

long sx, sy, tr, tg, n;
int minX, maxX, minY, maxY;
int w, h, largestc;

//  Due to a bug in C++ this program cannot be compiled with
// the optimization Automatic Register Variables.

//  Global Procedures

void far StatusFunc (int PercentDone)
  {
  if (PercentDone==99) PercentDone++;
  if ((PercentDone != LastPercent) && (PercentDone >= 0) && (!(PercentDone%5)))
    {
    char c[40];
    itoa (PercentDone, c, 10);
    strcat (c, "% Done");
    StatText -> SetText (c);
    LastPercent = PercentDone;
    };
  MSG Msg;
  if (PeekMessage (&Msg, NULL, 0, 0, PM_REMOVE))
    {
    TranslateMessage (&Msg);
    DispatchMessage (&Msg);
    };
  };

inline int max (int i, int j) { return (i>j ? i : j); };

int CompareIntensity (int x, int y, int &smallc)
  {
  int r, g, b, result;
  image -> GetPixelTC (x, y, r, g, b);
  mask -> GetPixelGray (x, y, b);
  result = ((max(r, g) <= smallc+variation) && (!b) && (max (r, g) >= thres));
  if ((max (r, g) < smallc) && result) smallc = max(r, g);
  return result;
  };

int DoLine (int x, int y, int dir)
  {
  int x1=x+1, x2=x, x3, x4, y1;
  int done=0, size;
  int r, g, b;
  int largec, smallc, largex = x;

  image->GetPixelTC (largex, y, r, g, b);  smallc = max(r, g);
  while (((CompareIntensity (x1-1, y, smallc)) && (x1 > 1)) ||
         ((CompareIntensity (x1-2, y, smallc)) && (x1 > 1)))
    {
    x1--;
    mask -> PutPixelGray (x1, y, 255);
    };
  image->GetPixelTC (largex, y, r, g, b);  smallc = max(r, g);
  while (((CompareIntensity (x2+1, y, smallc)) && (x2<w-1)) ||
         ((CompareIntensity (x2+2, y, smallc)) && (x2<w-1)))
    {
    x2++;
    mask -> PutPixelGray (x2, y, 255);
    };
  size = x2-x1+1;
  if (x1 < minX) minX = x1;
  if (x2 > maxX) maxX = x2;
  do {
    y1 = y;  (dir ? y++ : y--);
    x = largex;
    image->GetPixelTC (largex, y1, r, g, b);  smallc = max(r, g);
    while ((!CompareIntensity (x, y, smallc)) && (x >= x1-1)) x--;
    if (x < x1-1)
      while ((!CompareIntensity (x, y, smallc)) && (x <= x2+1)) x++;
    if ((y > 0) && (y < h-1) && (x > x2+1) &&
      (CompareIntensity (largex, (dir ? y+1:y-1), smallc))) x = largex;
    if (x <= x2+1)
      {
      largex = x;
      if (y < minY) minY=y;
      if (y > maxY) maxY=y;
      x3 = x+1;  x4=x;
      image->GetPixelTC (largex, y, r, g, b);
      largec = smallc = max(r, g);
      while (((CompareIntensity (x3-1, y, smallc)) && (x3 > 1)) ||
             ((CompareIntensity (x3-2, y, smallc)) && (x3 > 1)))
	{
	x3--;
	mask -> PutPixelGray (x3, y, 255);
	image -> GetPixelTC (x3, y, r, g, b);
	if (max (r, g) > largec) largec = max (r, g), largex = x3;
	if (largec < largestc) largestc = largec;
	};
      image->GetPixelTC (largex, y, r, g, b);  smallc = max(r, g);
      while (((CompareIntensity (x4+1, y, smallc)) && (x4 < w-1)) ||
             ((CompareIntensity (x4+2, y, smallc)) && (x4 < w-1)))
	{
	x4++;
	mask -> PutPixelGray (x4, y, 255);
	image -> GetPixelTC (x4, y, r, g, b);
	if (max (r, g) > largec) largec = max (r, g), largex = x4;
	if (largec < largestc) largestc = largec;
	};
      size += x4-x3+1;
/*
      image->GetPixelTC (x3, y, r, g, b);  smallc = max(r, g);
      for (x=x3-2; x>=x1; x--)
	if (CompareIntensity (x, y, smallc)) size += DoLine (x, y, dir);
      image->GetPixelTC (x4, y, r, g, b);  smallc = max(r, g);
      for (x=x4+2; x<=x2; x++)
	if (CompareIntensity (x, y, smallc)) size += DoLine (x, y, dir);
      image->GetPixelTC (x1, y1, r, g, b);  smallc = max(r, g);
      for (x=x1-2; x>=x3; x--)
	if (CompareIntensity (x, y1, smallc)) size += DoLine (x, y1, !dir);
      image->GetPixelTC (x2, y1, r, g, b);  smallc = max(r, g);
      for (x=x2+2; x<=x4; x++)
        if (CompareIntensity (x, y1, smallc)) size += DoLine (x, y1, !dir);
*/
      x1 = x3;  x2 = x4;
      if (x1 < minX) minX = x1;
      if (x2 > maxX) maxX = x2;
      }
      else done = -1;
    if (dir && (y==h-1)) done = -1;
    if (!dir && (y==0)) done = -1;
    } while (!done);
  return size;
  };

void Fill (int x, int y)
  {
  int r, g, b;
  if (x < 1) x = 1;
  if (y < 1) y = 1;
  if (x > w-2) x = w-2;
  if (y > h-3) y = h-3;
  minX = maxX = x;  minY = y;  maxY = y+1;
  image -> GetPixelTC (x, y, r, g, b);  
  b = largestc = max (r, g);
  DoLine (x, y, 0);
  largestc = b;
  if (y+2 < h)
    {
    if (CompareIntensity (x, y+1, b)) DoLine (x, y+1, 1);
    else if (CompareIntensity (x-1, y+1, b)) DoLine (x-1, y+1, 1);
    else if (CompareIntensity (x+1, y+1, b)) DoLine (x+1, y+1, 1);
    else if (CompareIntensity (x, y+2, b)) DoLine (x, y+1, 1);
    else if (CompareIntensity (x-1, y+2, b)) DoLine (x-1, y+1, 1);
    else if (CompareIntensity (x+1, y+2, b)) DoLine (x+1, y+1, 1);
    };
  if (minX < 1) minX = 1;
  if (maxX < 1) maxX = 1;
  if (minY < 0) minY = 0;
  if (maxX > w-2) maxX = w-2;
  if (minX > w-2) minX = w-2;
  if (maxY > h-2) maxY = h-2;
  if (minY > h-2) minY = h-2;
  };

//  TMainWindow methods

TMainWindow::TMainWindow (PTWindowsObject AParent) :
    TWindow (AParent, "Cartlage Image Analysis")
  {
  Attr.W = 500;
  Attr.H = 150;
  AssignMenu (id_MAINMENU);
  FuncText = new TStatic (this, 98, "", 10, 10, 230, 24, 40);
  StatText = new TStatic (this, 99, "", 250, 10, 150, 24, 40);
  image = 0;  busy = 0;  pal = NULL;  nNames = 0;  wantToQuit = 0;
  DisplayWindow = NULL;
  MakePalette ();
  };

int TMainWindow::CanClose ()
  {
  wantToQuit = 1;
  return (!busy);
  };

void TMainWindow::ClearDisplay ()
  {
  DisplayWindow = NULL;
  };

void TMainWindow::CMBatch (RTMessage Msg)
  {
  if (busy) return;
  wantToQuit = 0;
  while ((nNames) && (!wantToQuit))
    {
    strcpy (ORGname, LISTnames[nNames-1]);
    strcpy (ENHname, ORGname);
      strcpy (strrchr (ENHname, '.'), ".TMP");
    strcpy (DATname, ORGname);
      strcpy (strrchr (DATname, '.'), ".DAT");
    strcpy (LINname, ORGname);
      strcpy (strrchr (LINname, '.'), ".LIN");
    CMEnhance (Msg);
    CMAnalize (Msg);
    nNames--;
    InvalidateRect (HWindow, NULL, 0);
    };
  };

void TMainWindow::CMBoundry (RTMessage Msg)
  {
  if (!strlen(ORGname))
    {
    MessageBox (HWindow, "No File Selected", "Error", MB_OK | MB_ICONEXCLAMATION);
    return;
    };
  if (busy) return;
  busy++;
  image = new MemoryImage (8, 8, 8);
  image->SetStatusFunction (&StatusFunc);

  FuncText->SetText ("Loading Image");
  TGAFile *inimage = new TGAFile (ORGname, (tAccess)READ);
  image->LoadImage (inimage);
  delete inimage;
  FuncText->SetText ("Displaying");
  Display ();

  FuncText->SetText ("Complete");
  StatText->SetText ("");
  delete image;
  busy--;
  };

void TMainWindow::CMOpen (RTMessage Msg)
  {
  if (busy) return;
  char FileName[256];
  if (GetApplication()->ExecDialog(new TFileDialog (this, SD_FILEOPEN, strcpy (FileName, "*.TGA"))) == IDOK)
    {
    strcpy (LISTnames[nNames], FileName);  nNames++;
    strcpy (ORGname, FileName);
    strcpy (ENHname, FileName);
      strcpy (strrchr (ENHname, '.'), ".TMP");
    strcpy (DATname, FileName);
      strcpy (strrchr (DATname, '.'), ".DAT");
    strcpy (LINname, FileName);
      strcpy (strrchr (LINname, '.'), ".LIN");
    InvalidateRect (HWindow, NULL, 0);
    } else strcpy (ORGname, "");
  };

void TMainWindow::Display ()
  {
  if (!DisplayWindow)
    {
    int w, h;
    DisplayWindow = new TDisplayWindow (this);
    image->GetImageSize (w, h);
    w += 2 * GetSystemMetrics (SM_CXFRAME);
    h += 2 * GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION)
        /* + GetSystemMetrics(SM_CYMENU) */ ;
    DisplayWindow->Attr.W = w;
    DisplayWindow->Attr.H = h;
    GetApplication()-> MakeWindow (DisplayWindow);
    };
  DisplayWindow->SetPalette (pal);
  DisplayWindow->Display();
  };

void TMainWindow::MakePalette()
  {
  int i;
  LPLOGPALETTE Palette = (LPLOGPALETTE) malloc (sizeof(LOGPALETTE) + sizeof(PALETTEENTRY)*128);
  if (!Palette) return;
  Palette->palVersion = 0x300;
  Palette->palNumEntries = 128;
  for (i=0; i<128; i++)
    Palette->palPalEntry[i].peRed = (i&0x07)<<5,
    Palette->palPalEntry[i].peGreen = (i>>3)<<4,
    Palette->palPalEntry[i].peBlue = 0,
    Palette->palPalEntry[i].peFlags = PC_RESERVED;
  pal = CreatePalette (Palette);
  free ((void *)Palette);
  if (!pal) return;
  };

void TMainWindow::CMEnhance (RTMessage Msg)
  {
  if (!strlen(ORGname))
    {
    MessageBox (HWindow, "No File Selected", "Error", MB_OK | MB_ICONEXCLAMATION);
    return;
    };
  if (busy) return;
  busy++;
  image = new MemoryImage (8, 8, 8);
  image->SetStatusFunction (&StatusFunc);

  FuncText->SetText ("Loading Image");
  TGAFile *inimage = new TGAFile (ORGname, (tAccess)READ);
  if (DisplayWindow) SendMessage (DisplayWindow->HWindow, WM_CLOSE, 0, 0);
  image->LoadImage (inimage);
  delete inimage;
  FuncText->SetText ("Displaying");
  Display ();

  FuncText->SetText ("Smoothing");
  image->Smooth (green, 5);
  image->Smooth (red, 5);
  image->Smooth (green, 5);
  image->Smooth (red, 5);
  FuncText->SetText ("Displaying");
  Display ();

  FuncText->SetText ("Calculating Histogram");
  image->CalculateHistogram();
  image->SmoothHistogram();

  FuncText->SetText ("Contrast Enhancement");
  image->EnhanceContrast (red, image->GetBackgroundIntensity (red),
			       image->GetMaximumIntensity (red));
  image->EnhanceContrast (green, image->GetBackgroundIntensity (green),
				 image->GetMaximumIntensity (green));
  FuncText->SetText ("Shifting Intensities");
  image->ShiftIntensity (red, 1);
  FuncText->SetText ("Smoothing");
  image->Smooth (green, 3);
  image->Smooth (red, 3);
  FuncText->SetText ("Displaying");
  Display ();

  FuncText->SetText ("Saving Image");
  HCURSOR oldCursor = SetCursor (LoadCursor (NULL, IDC_WAIT));
  TGAFile *outimage = new TGAFile (ENHname, (tAccess)WRITE);
  outimage -> SetPixelSize (24, 1);
  outimage -> LoadImage (image);
  delete outimage;
  SetCursor (oldCursor);
  FuncText->SetText ("Complete");
  StatText->SetText ("");
  MessageBeep (0);
  busy--;
  delete image;
  };

void TMainWindow::CMAnalize (RTMessage Msg)
  {
  if (!strlen(ORGname))
    {
    MessageBox (HWindow, "No File Selected", "Error", MB_OK | MB_ICONEXCLAMATION);
    return;
    };
  if (busy) return;
  busy++;  numcells = 0;
  f = new ofstream (DATname);
  *f << "x" << "\t" << "y" << "\t";
  *f << "tRed" << "\t" << "tGreen" << "\t" << "pixels" << "\n";

  image = new MemoryImage (8, 8, 8);
  image->SetStatusFunction (&StatusFunc);
 
  TGAFile *inimage = new TGAFile (ENHname, (tAccess)READ);
  FuncText->SetText ("Loading Image");
  image->LoadImage (inimage);
  delete inimage;
  FuncText->SetText ("Displaying");
  Display ();

  FuncText->SetText ("Counting Cells");
  StatText->SetText ("");
  int x, y, c, i;
  image -> GetImageSize (w, h);
  mask = new MemoryImage (1);
  if (mask -> CreateImage (w, h)) return;
  mask -> ClearImage (gray, 0);
  for (x=0; x<w; x++)
    {
    image->PutPixelTC (x, 0, 1, 0, 0);
    image->PutPixelTC (x, h-1, 1, 0, 0);
    };
  for (y=0; y<h; y++)
    {
    image->PutPixelTC (0, y, 1, 0, 0),
    image->PutPixelTC (w-1, y, 1, 0, 0);
    };
  x = 0;  y = 0;
  do {
    FindMax (x, y, (x+blocksize < w ? x+blocksize : w-1),
	       (y+blocksize < h ? y+blocksize : h-1));
    x += blocksize;
    if (x>=w) x=0, y+=blocksize;
    } while (y<h);
  FuncText->SetText ("Displaying");
  Display ();
  delete f;
  FuncText->SetText ("Complete");
  StatText->SetText ("");
  MessageBeep (0);  busy--;
  delete image;
  };

void TMainWindow::FindMax (int x1, int y1, int x2, int y2)
  {
  int mx, my, x, y;
  int r, g, b;
  int c;

  if (x2-x1 < blocksize) x1 = x2-blocksize;
  if (y2-y1 < blocksize) y1 = y2-blocksize;
  if (x1 < 0) x2 -= x1, x1 = 0;
  if (y1 < 0) y2 -= y1, y1 = 0;
  do {
    c = -1;
    for (y=y1; y<=y2; y++)
      {
      StatusFunc (-1);
      for (x=x1; x<=x2; x++)
	{
	image -> GetPixelTC (x, y, r, g, b);
	if (max (r, g) > c)
	  {
	  c = max (r, g);
	  mx = x;  my = y;
	  };
	};
      };
    if (c > thres)
      {
      if ((mx > x1+edgesize) && (mx < x2-edgesize)
	  && (my > y1+edgesize) && (my < y2-edgesize))
	FindObject (mx, my, c);
      else
	if (((mx<=edgesize) && (x1==0)) || ((my<=edgesize) && (y1==0)) ||
	   ((mx>=w-edgesize-1) && (x2==w-1)) || ((my>=h-edgesize-1) && (y2==h-1)))
	  FindObject (mx, my, c);
	else FindMax ((mx>nextBlockSize ? mx-nextBlockSize : 0),
		      (my>nextBlockSize ? my-nextBlockSize : 0),
		      (mx+nextBlockSize<w ? mx+nextBlockSize : w-1),
		      (my+nextBlockSize<h ? my+nextBlockSize : h-1));
      };
    } while (c > thres);
  };

void TMainWindow::FindObject (int x, int y, int c)
  {
  DisplayWindow->Circle (x, y);
  char ch[40];  numcells++;
  itoa (numcells, ch, 10);
  strcat (ch, " Done");
  StatText -> SetText (ch);

  int c1;
  sx=sy=tr=tg=n = 0;
  int r, g, b;
  Fill (x, y);
  mask -> PutPixelGray (x, y, 255);
  StatusFunc (-1);
  for (y=minY; y<=maxY+1; y++)
    for (x=minX-1; x<=maxX+1; x++)
      {
      mask -> GetPixelGray (x, y, c1);
      if (c1)
        {
	mask -> PutPixelGray (x, y, 0);
	image -> GetPixelTC (x, y, r, g, b);
	image -> PutPixelTC (x, y, 0, 0, 0);
	if (max(r, g) > c-countRange)
	  {
	  tr += r;  tg += g;  n++;
	  sx += x*(long)(r+g);  sy += y*(long)(r+g);
	  };
	};
      };
  sx /= tr + tg;  sy /= tr + tg;
  image -> PutPixelTC (sx, sy, thres-2, thres-2, 255);
  *f << sx << "\t" << sy << "\t";
  *f << tr << "\t" << tg << "\t" << n << "\n";
  };

void TMainWindow::Paint (HDC PaintDC, PAINTSTRUCT _FAR & PaintInfo)
  {
  int i;
  for (i=0; i<nNames; i++)
    if (i<20)
      TextOut (PaintDC, 10, i*20+35, LISTnames[i], strlen(LISTnames[i]));
    else
      TextOut (PaintDC, 100, (i-20)*20+35, LISTnames[i], strlen(LISTnames[i]));
  };

//  TDisplayWindow Methods

TDisplayWindow::TDisplayWindow (PTWindowsObject AParent) :
    TWindow (AParent, "Displayed Image")
  {
  Attr.Style |= WS_OVERLAPPEDWINDOW | WS_POPUPWINDOW;
  Attr.X = 100;  Attr.Y = 100;
  Attr.W = 200;  Attr.H = 200;
  memDC = NULL;    pal = NULL;  displaying = 1;  oldBM = NULL;  dragDC = NULL;
  topline.numpoints = 0;
  boneline.numpoints = 0;
  s_iline.numpoints = 0;
  i_dline.numpoints = 0;
  leftline.numpoints = 0;
  rightline.numpoints = 0;
  curline.numpoints = 0;
  };

TDisplayWindow::~TDisplayWindow ()
  {
  if (memDC)
    {
    DeleteObject (SelectObject (memDC, oldBM));
    DeleteDC (memDC);
    };
  };

int TDisplayWindow::CanClose ()
  {
  ((TMainWindow *)Parent)->ClearDisplay();
  return 1;
  };

void TDisplayWindow::Circle (int x, int y)
  {
  LOGBRUSH brush;
  brush.lbStyle = BS_HOLLOW;
  HGDIOBJ oldBrush = SelectObject (memDC, CreateBrushIndirect(&brush));
  HGDIOBJ oldPen = SelectObject (memDC, CreatePen (PS_SOLID, 1, RGB (0xFF, 0xFF, 0xFF)));
  Ellipse (memDC, x-5, y-5, x+5, y+5);
  DeleteObject (SelectObject (memDC, oldPen));
  DeleteObject (SelectObject (memDC, oldBrush));
  InvalidateRect (HWindow, NULL, FALSE);
  };

void TDisplayWindow::Display ()
  {
  displaying = 1;
  HDC dc = GetDC (HWindow);
  if (memDC)
    {
    DeleteObject (SelectObject (memDC, oldBM));
    DeleteDC (memDC);
    };
  memDC = CreateCompatibleDC (dc);

  image->GetImageSize (w, h);

  SelectPalette (dc, pal, 0);
  RealizePalette (dc);

  HBITMAP memBM = (HBITMAP) image->CreateBitmap ((unsigned int) dc);
  oldBM = (HBITMAP) SelectObject (memDC, memBM);

  BitBlt (dc, imageXofs, imageYofs, w, h, memDC, 0, 0, SRCCOPY);
  ReleaseDC (HWindow, dc);
  displaying = 0;
  };

void TDisplayWindow::Paint (HDC PaintDC, PAINTSTRUCT _FAR &PaintInfo)
  {
  if ((memDC) && (!displaying))
    {
    if (pal) SelectPalette (PaintDC, pal, 0);
    RealizePalette (PaintDC);
    BitBlt (PaintDC, imageXofs, imageYofs, w, h, memDC, 0, 0, SRCCOPY);
    };
  int i;
  LOGBRUSH brush;
  brush.lbStyle = BS_HOLLOW;
  HGDIOBJ oldBrush = SelectObject (PaintDC, CreateBrushIndirect(&brush));
  HGDIOBJ oldPen = SelectObject (PaintDC, CreatePen (PS_SOLID, 1, RGB (0xFF, 0xFF, 0)));
  if (topline.numpoints)
    {
    MoveTo (PaintDC, topline.points[0].x, topline.points[0].y);
    for (i=1; i<topline.numpoints; i++)
      LineTo (PaintDC, topline.points[i].x, topline.points[i].y);
    };
  if (s_iline.numpoints)
    {
    MoveTo (PaintDC, s_iline.points[0].x, s_iline.points[0].y);
    for (i=1; i<s_iline.numpoints; i++)
      LineTo (PaintDC, s_iline.points[i].x, s_iline.points[i].y);
    };
  if (i_dline.numpoints)
    {
    MoveTo (PaintDC, i_dline.points[0].x, i_dline.points[0].y);
    for (i=1; i<i_dline.numpoints; i++)
      LineTo (PaintDC, i_dline.points[i].x, i_dline.points[i].y);
    };
  if (boneline.numpoints)
    {
    MoveTo (PaintDC, boneline.points[0].x, boneline.points[0].y);
    for (i=1; i<boneline.numpoints; i++)
      LineTo (PaintDC, boneline.points[i].x, boneline.points[i].y);
    };
  if (leftline.numpoints)
    {
    MoveTo (PaintDC, leftline.points[0].x, leftline.points[0].y);
    for (i=1; i<leftline.numpoints; i++)
      LineTo (PaintDC, leftline.points[i].x, leftline.points[i].y);
    };
  if (rightline.numpoints)
    {
    MoveTo (PaintDC, rightline.points[0].x, rightline.points[0].y);
    for (i=1; i<rightline.numpoints; i++)
      LineTo (PaintDC, rightline.points[i].x, rightline.points[i].y);
    };
  DeleteObject (SelectObject (PaintDC, oldPen));
  DeleteObject (SelectObject (PaintDC, oldBrush));
  };

void TDisplayWindow::SaveLines ()
  {
  ofstream *out = new ofstream (LINname);
  int i;

  *out << "TOP LINE" << "\n";
  for (i=0; i<topline.numpoints; i++)
    *out << topline.points[i].x << "\t" << topline.points[i].y << "\n";
  *out << "SUP-INT LINE" << "\n";
  for (i=0; i<s_iline.numpoints; i++)
    *out << s_iline.points[i].x << "\t" << s_iline.points[i].y << "\n";
  *out << "INT-DEEP LINE" << "\n";
  for (i=0; i<i_dline.numpoints; i++)
    *out << i_dline.points[i].x << "\t" << i_dline.points[i].y << "\n";
  *out << "BONE LINE" << "\n";
  for (i=0; i<boneline.numpoints; i++)
    *out << boneline.points[i].x << "\t" << boneline.points[i].y << "\n";
  *out << "LEFT LINE" << "\n";
  for (i=0; i<leftline.numpoints; i++)
    *out << leftline.points[i].x << "\t" << leftline.points[i].y << "\n";
  *out << "RIGHT LINE" << "\n";
  for (i=0; i<rightline.numpoints; i++)
    *out << rightline.points[i].x << "\t" << rightline.points[i].y << "\n";

  delete out;
  };

inline void TDisplayWindow::SetPalette (HPALETTE p) { pal = p; };

void TDisplayWindow::WMLButtonDown (RTMessage Msg)
  {
  if (!dragDC)
    {
    SetCapture (HWindow);
    dragDC = GetDC (HWindow);
    MoveTo (dragDC, Msg.LP.Lo, Msg.LP.Hi);
    curline.points[0].x = Msg.LP.Lo;
    curline.points[0].y = Msg.LP.Hi;
    curline.numpoints = 1;
    };
  };

void TDisplayWindow::WMMouseMove (RTMessage Msg)
  {
  if (dragDC)
    {
    LOGBRUSH brush;
    brush.lbStyle = BS_HOLLOW;
	 HGDIOBJ oldBrush = SelectObject (dragDC, CreateBrushIndirect(&brush));
	 HGDIOBJ oldPen = SelectObject (dragDC, CreatePen (PS_SOLID, 1, RGB (0, 0xFF, 0xFF)));
    LineTo (dragDC, Msg.LP.Lo, Msg.LP.Hi);
    int n = curline.numpoints;
    if ((n>1) && ((Msg.LP.Hi-curline.points[n-1].y)*(Msg.LP.Lo-curline.points[n-2].x)
    		==(Msg.LP.Hi-curline.points[n-2].y)*(Msg.LP.Lo-curline.points[n-1].x)))
      curline.numpoints--;
    curline.points[curline.numpoints].x = Msg.LP.Lo;
    curline.points[curline.numpoints].y = Msg.LP.Hi;
    curline.numpoints++;
    if (curline.numpoints >= maxpoints) curline.numpoints--;
    DeleteObject (SelectObject (dragDC, oldPen));
    DeleteObject (SelectObject (dragDC, oldBrush));
	 };
  };

void TDisplayWindow::WMLButtonUp (RTMessage Msg)
  {
  if (dragDC)
    {
    ReleaseCapture ();
    ReleaseDC (HWindow, dragDC);  dragDC = NULL;

    TDialog *p;
    selectlinemsg m;
    p = new TDialog (this, id_SELECTLINE, NULL);
    p->SetTransferBuffer (&m);
    new TRadioButton (p, 101, NULL);  m.top   = 1;
    new TRadioButton (p, 102, NULL);  m.s_i   = 0;
    new TRadioButton (p, 103, NULL);  m.i_d   = 0;
    new TRadioButton (p, 104, NULL);  m.deep  = 0;
    new TRadioButton (p, 105, NULL);  m.left  = 0;
    new TRadioButton (p, 106, NULL);  m.right = 0;
    if (GetApplication()->ExecDialog(p) == IDOK)
      {
      if (m.top)    topline = curline;
      if (m.s_i)    s_iline = curline;
      if (m.i_d)    i_dline = curline;
      if (m.deep)   boneline = curline;
      if (m.left)   leftline = curline;
      if (m.right)  rightline = curline; 
      };
    InvalidateRect (HWindow, NULL, FALSE);
    SaveLines ();
    };
  };

//  TApp methods

void TApp::InitMainWindow ()
  {
  MainWindow = new TMainWindow (NULL);
  };

// Main Program

int PASCAL WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
  {
  TApp App ("ImageApp", hInstance, hPrevInstance, lpCmdLine, nCmdShow);
  App.Run();
  return App.Status;
  };
